"""
Paicos is an object-oriented Python package for analysis of (cosmological)
simulations performed with Arepo. Its main feature is that it includes
handling of units, derived variables, loading of data upon request.

The module includes a way of writing and reading data with units.
The code is parallel with an OpenMP Cython implementation
and a CUDA GPU implementation for visualization.
"""

__version__ = "0.1.14"
__author__ = 'Thomas Berlok'
__credits__ = 'Niels Bohr Institute, University of Copenhagen'

# Dependencies
import os
import numpy
import scipy
import h5py
import astropy

# Settings and utility functions
from . import util
from . import settings

# One folder up from __init__ (i.e. repo directory or installation directory)
from .util import root_dir

# HDF5 file readers
from .readers.arepo_snap import Snapshot
from .readers.arepo_catalog import Catalog
from .readers.paicos_readers import PaicosReader, ImageReader, Histogram2DReader
from .readers.generic_snap import GenericSnapshot


# HDF5 file writers
from .writers.paicos_writer import PaicosWriter, PaicosTimeSeriesWriter
from .writers.arepo_image import ImageWriter, ArepoImage

# Image creators
from .image_creators.image_creator import ImageCreator
from .image_creators.projector import Projector
from .image_creators.nested_projector import NestedProjector
from .image_creators.tree_projector import TreeProjector
from .image_creators.slicer import Slicer

# Histograms
from .histograms.histogram import Histogram
from .histograms.histogram2D import Histogram2D

# Derived variables
from .derived_variables import derived_variables

# Orientation class
from .orientation import Orientation

# Cython functions
from . import cython

# pylint: disable=W0621

# The place where __init__.py (this file) is located
code_dir = os.path.dirname(os.path.abspath(__file__))
# This is the user home_dir
home_dir = os.path.expanduser('~')


def use_units(use_units):
    """
    Parameters
    ----------

    use_units : bool
        True for enabling units globally and False for disabling
        them globally.
        Note that this will not modify snapshots that have
        already been loaded.

    Examples
    --------

    An example::

        import paicos as pa
        print('Using units:', pa.settings.use_units)

        # Turn on paicos quantities globally
        pa.use_units(True)
        print('Using units:', pa.settings.use_units)

        # Loads in data without applying units.
        pa.use_units(False)
        print('Using units:', pa.settings.use_units)

    """
    settings.use_units = use_units


def add_user_function(variable_string, function):
    """
    This functions allows to enable user functions for obtaining
    deriving variables. An example could be the following::

        def TemperaturesTimesMassesSquared(snap, get_depencies=False):
            if get_depencies:
                return ['0_Temperatures', '0_Masses']
            return snap['0_Temperatures'] * snap['0_Masses']**2


        pa.add_user_function('0_TM2', TemperaturesTimesMassesSquared)

    """
    from inspect import signature
    if ' ' in variable_string:
        raise RuntimeError(f'{variable_string} contains space(s)!')

    sig = signature(function)
    if len(sig.parameters) == 2:
        dependencies = function(None, True)
        assert isinstance(dependencies, list)
        err_msg = 'Problem with list of dependencies!'
        for item in dependencies:
            assert isinstance(item, str), err_msg
            assert ' ' not in item, err_msg

    derived_variables.user_functions.update({variable_string: function})


def use_only_user_functions(use_only_user_functions):
    """
    Passing 'True' to this functions disables automatic derivation of
    variables except for the ones you have added via 'add_user_function'.
    """
    settings.use_only_user_functions = use_only_user_functions


def add_user_unit(field, blockname, unit):
    """
    This function adds user units.

    Parameters
    ----------

    field : string
        possible values are specified in
        unit_specifications.pos_fields, currently

        pos_fields = ['default', 'voronoi_cells', 'dark_matter',
        'stars', 'black_holes', 'groups', 'subhalos']

    blockname : string
        The blockname that you would like to enable units for.

    unit : string
        The unit that the blockname should have.
        String inputs are preferred, e.g., ``arepo_mass arepo_length small_a^2 small_h^(-3/2)``.

        The available arepo units are:

        ``arepo_length`` (often 1 kpc)

        ``arepo_mass`` (often 10^10 Msun)

        ``arepo_velocity`` (often 1 km/s)

        ``arepo_time``

        ``arepo_energy``

        ``arepo_pressure``

        ``arepo_density``

        Please note that these only become available once you
        actually load a hdf5 file. This is the reason that you can
        pass the string only.
    """
    if field not in util.user_unit_dict:
        raise RuntimeError(f'unknown field: {field}')

    util.user_unit_dict[field][blockname] = unit


def numthreads(numthreads):
    """
    Pass the number of threads you would like to use for parallel execution.

    numthreads (int): e.g. 16
    """
    util._check_if_omp_has_issues(False)

    if numthreads > settings.max_threads:
        print(f'Your machine only has {settings.max_threads} available threads')

    settings.numthreads = min(numthreads, settings.max_threads)
    if settings.openMP_has_issues:
        settings.numthreads_reduction = 1
    else:
        settings.numthreads_reduction = settings.numthreads


def print_info_when_deriving_variables(option):
    """
    Input: a boolean controlling whether to provide info to terminal.
    """
    settings.print_info_when_deriving_variables = option


def give_openMP_warnings(option):
    """
    Turns of warnings that might otherwise appear on a mac computer.
    """
    settings.give_openMP_warnings = option


def load_cuda_functionality_on_startup(option):
    """
    Turns on/off whether to import GPU/cuda on startup.
    """
    settings.load_cuda_functionality_on_startup = option


def set_aliases(aliases):
    """
    Assign a list of aliases for use as keys in snapshot objects.

    Parameters
    ----------
        aliases : dict
            A dictionary containing the mapping of variable names
            that you would like to have enabled, e.g.::

                aliases = {'0_Density': 'dens',
                       '0_Temperatures': 'T',
                       '0_MeanMolecularWeight': 'mu'}
    """
    msg = 'Your user-set aliases seem to not be unique'
    assert len(set(aliases.values())) == len(aliases.values()), msg

    inverse = {aliases[key]: key for key in aliases.keys()}
    settings.aliases = aliases
    settings.inverse_aliases = inverse
    settings.use_aliases = True


def import_user_settings():
    """
    Import user settings if they exist in the root directory of Paicos
    or as a hidden file at the users home directory.

    :meta private:
    """
    data_dir = None
    if os.path.exists(root_dir + 'data/'):
        data_dir = root_dir + 'data/'
    if os.path.exists(code_dir + '/paicos_user_settings.py'):
        from . import paicos_user_settings
        if hasattr(paicos_user_settings, 'data_dir'):
            data_dir = paicos_user_settings.data_dir
    if os.path.exists(home_dir + '/.paicos_user_settings.py'):
        filepath = home_dir + '/.paicos_user_settings.py'
        import importlib
        spec = importlib.util.spec_from_file_location('paicos_settings',
                                                      location=filepath)
        foo = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(foo)
        if hasattr(foo, 'data_dir'):
            data_dir = foo.data_dir

    if data_dir is not None:
        if data_dir[-1] != '/':
            data_dir += '/'

    return data_dir


data_dir = import_user_settings()


# Import of GPU functionality only if
# a simple test with cupy and numba works.
def gpu_init(gpu_num=0):
    """
    Calling this function initializes the GPU parts of the code.

    You can set settings.load_cuda_functionality_on_startup = True,
    to do this automatically on startup, but you should be aware
    that loading cupy and numba can be a bit time-consuming,
    so that this significantly affects the import time.

    Parameters:
    ----------

        gpu_num (int): The GPU that you want to use for computations,
                       i.e., we call cupy.cuda.Device(gpu_num).use()
    """

    try:
        import cupy as cp
        from numba import cuda

        if gpu_num != 0:
            cp.cuda.Device(gpu_num).use()

        @cuda.jit
        def my_kernel(io_array):
            pos = cuda.grid(1)
            if pos < io_array.size:
                io_array[pos] *= 2

        data = cp.ones(10**6)
        threadsperblock = 256
        blockspergrid = (data.size + (threadsperblock - 1)) // threadsperblock
        my_kernel[blockspergrid, threadsperblock](data)

        del data

        # Test above worked, do the imports
        from .image_creators.gpu_sph_projector import GpuSphProjector
        from .image_creators.gpu_ray_projector import GpuRayProjector
        import paicos
        paicos.GpuSphProjector = GpuSphProjector
        paicos.GpuRayProjector = GpuRayProjector
    except Exception as e:
        import warnings
        print(e)
        err_msg = ('\nPaicos: The simple cuda example using cupy and numba failed '
                   'with the error above. Please check the official documentation for '
                   'cupy and numba for installation procedure. Note that you need '
                   'a GPU that supports CUDA.\n')
        warnings.warn(err_msg)


if settings.load_cuda_functionality_on_startup:
    gpu_init()


# Do this at start up
util._check_if_omp_has_issues()



"""
This module defines a class for 2D histograms.
"""
import numpy as np
import h5py
from astropy import units as u
from .. import units as pu
from .. import util
from .. import settings
from .histogram import _make_bins
from ..cython.histogram import find_normalizing_norm_of_2d_hist


class Histogram2D:
    """
    This code defines a Histogram2D class which can be used to create 2D
    histograms. The class takes in the bin edges for the x and y axes, and an
    optional argument to indicate if the histogram should be in log scale. The
    class has methods to calculate the bin edges and centers, remove astro
    units, and create the histogram with a specific normalization. It also has
    a method to generate a color label for the histogram with units.
    """

    def __init__(self, snap, x, y, weights=None, bins_x=200, bins_y=200,
                 normalize=True, logscale=True):
        """
        Initialize the Histogram2D class with the bin edges for the x
        and y axes, and an optional argument to indicate if the
        histogram should be in log scale.

        Parameters:
            snap (Snapshot): the input snapshot

            x (array): The x data for the histogram

            y (array): The y data for the histogram

            weights (array): The weight data for the histogram, default
                             is None

            bins_x (tuple): Tuple of lower edge, upper edge and number
                            of bins for x axis. Alternatively an integer
                            denoting the number of bins spanning
                            x.min() to x.max().

            bins_y (tuple): Tuple of lower edge, upper edge and number
                            of bins for y axis. Alternatively an integer.

            normalize (bool): Indicates whether the histogram should be
                               normalized, default is True

            logscale (bool): Indicates whether to use logscale for the
                             histogram, default is True.


        """

        self.snap = snap

        if isinstance(x, str):
            self.x = snap[x]
        else:
            self.x = x

        if isinstance(y, str):
            self.y = snap[y]
        else:
            self.y = y

        if isinstance(weights, str):
            self.weights = snap[weights]
        else:
            self.weights = weights

        self.normalize = normalize

        if isinstance(bins_x, int):
            bins_x = [self.x.min(), self.x.max(), bins_x]

            if logscale:
                bins_x[0] = self.x[self.x > 0].min()
        else:
            assert bins_x[0] < bins_x[1], 'min and max values swapped!'

        if isinstance(bins_y, int):
            bins_y = [self.y.min(), self.y.max(), bins_y]

            if logscale:
                bins_y[0] = self.y[self.y > 0].min()
        else:
            assert bins_y[0] < bins_y[1], 'min and max values swapped!'

        self.logscale = logscale

        if logscale:
            assert bins_x[0] > 0
            assert bins_y[0] > 0

        self.edges_x, self.centers_x = _make_bins(bins_x, self.logscale)
        self.edges_y, self.centers_y = _make_bins(bins_y, self.logscale)
        self.lower_x = self.edges_x[0]
        self.lower_y = self.edges_y[0]
        self.upper_x = self.edges_x[-1]
        self.upper_y = self.edges_y[-1]

        self.extent = [self.lower_x, self.upper_x, self.lower_y, self.upper_y]

        self._get_image_properties()

        # Make the histogram
        self.hist2d = self._make_histogram()

        self.colorlabel = None

    def _get_image_properties(self):
        dx = np.diff(self.edges_x)
        dy = np.diff(self.edges_y)

        if settings.use_units:
            dxu = dx.unit_quantity
            dyu = dy.unit_quantity
            dxx, dyy = np.meshgrid(dx.value, dy.value)
            dxx = dxx * dxu
            dyy = dyy * dyu

            centers_x_mat, centers_y_mat = np.meshgrid(self.centers_x.value,
                                                       self.centers_y.value)
            centers_x_mat = centers_x_mat * dxu
            centers_y_mat = centers_y_mat * dyu
        else:
            dxx, dyy = np.meshgrid(dx, dy)
            centers_x_mat, centers_y_mat = np.meshgrid(self.centers_x,
                                                       self.centers_y)

        self.area_per_bin = dxx * dyy

        if self.logscale and settings.use_units:
            self.area_per_bin *= u.Unit('dex')**(2) / self.area_per_bin.unit
        self.centers_x_mat = centers_x_mat
        self.centers_y_mat = centers_y_mat

    def get_colorlabel(self, x_symbol, y_symbol, weight_symbol=None):
        """
        Method to generate a color label for the histogram with units

        Parameters:
            x_symbol (string): Symbol for x axis
            y_symbol (string): Symbol for y axis
            weight_symbol (string): Symbol for weight of the histogram,
                                    default is None

        Returns:
            string: The color label for the histogram with units
        """

        assert settings.use_units

        unit_label = self.hist2d.label()[1:-1]

        if self.logscale:
            colorlabel = (r'/\left(\mathrm{d}\mathrm{log}_{10} ' + x_symbol
                          + r'\,\mathrm{d}\mathrm{log}_{10}'
                          + y_symbol + r'\right)'
                          + r'\;' + unit_label)
        else:
            colorlabel = (r'/\left(\mathrm{d}' + x_symbol
                          + r'\,\mathrm{d}' + y_symbol + r'\right)'
                          + r'\;' + unit_label)

        if weight_symbol is not None:
            if self.normalize:
                colorlabel = weight_symbol + \
                    r'_\mathrm{tot}^{-1}\,\mathrm{d}' + \
                    weight_symbol + colorlabel
            else:
                colorlabel = r'\mathrm{d}' + weight_symbol + colorlabel
        else:
            if self.normalize:
                colorlabel = r'\mathrm{pixel count}/\mathrm{total count}' + colorlabel
            else:
                colorlabel = r'\mathrm{pixel count}' + colorlabel
        self.colorlabel = r'$' + colorlabel + r'$'
        return self.colorlabel

    @util.remove_astro_units
    def _cython_make_histogram(self, x, y, edges_x, edges_y, weights):
        """
        Private method for making the 2D histogram using cython code
        """

        if settings.openMP_has_issues:
            from ..cython.histogram import get_hist2d_from_weights as get_hist2d
        else:
            from ..cython.histogram import get_hist2d_from_weights_omp as get_hist2d

        nbins_x = edges_x.shape[0] - 1
        nbins_y = edges_y.shape[0] - 1

        lower_x = edges_x[0]
        upper_x = edges_x[-1]
        lower_y = edges_y[0]
        upper_y = edges_y[-1]

        hist2d = get_hist2d(
            x, y, weights,
            lower_x, upper_x, nbins_x,
            lower_y, upper_y, nbins_y,
            self.logscale,
            numthreads=settings.numthreads_reduction)

        return hist2d

    def _make_histogram(self):
        """
        Private method to create the 2D histogram.

        Returns:
            hist2d (2D array): The 2D histogram
        """

        x = self.x
        y = self.y
        weights = self.weights
        normalize = self.normalize

        # Figure out units for the histogram
        if settings.use_units:
            if not normalize and (weights is not None):
                hist_units = weights.unit
            else:
                hist_units = u.Unit('')
            if self.logscale:
                hist_units *= u.Unit('dex')**(-2)
            else:
                hist_units /= x.unit * y.unit

            self.hist_units = hist_units

            assert x.unit == self.edges_x.unit
            assert y.unit == self.edges_y.unit

        if weights is None:
            # np.ones_like returns something with units,
            # should replace with np.ones and the shape
            weights = np.ones(x.shape, dtype=np.float64)

        hist2d = self._cython_make_histogram(x, y, self.edges_x,
                                             self.edges_y, weights)

        hist2d = hist2d.T

        if settings.use_units:
            hist2d = pu.PaicosQuantity(hist2d, self.hist_units, a=self.x._a,
                                       h=self.x._h,
                                       comoving_sim=self.x.comoving_sim)
        if normalize:
            norm = np.sum(self.area_per_bin * hist2d)
            hist2d /= norm

            sanity = np.sum(self.area_per_bin * hist2d)
            if settings.use_units:
                np.testing.assert_allclose(sanity.value, 1.0)
                assert sanity.unit == u.Unit(''), f'{sanity.unit} should be dimensionless'
            else:
                np.testing.assert_allclose(sanity, 1.0)

        return hist2d

    def save(self, basedir, basename="2d_histogram"):
        """
        Saves the 2D histogram in the basedir directory.

        Parameters:
            basedir (path): The directory where the histogram should be saved.
            basename (string): The basename for the filename, which will take
                               the form::

                                   filename =  basename + f'_{snapnum:03d}.hdf5'
        """

        if basedir[-1] != '/':
            basedir += '/'

        snapnum = self.snap.snapnum
        filename = basedir + basename + f'_{snapnum:03d}.hdf5'
        with h5py.File(filename, 'w') as hdf5file:
            #
            hdf5file.create_group('hist_info')
            hdf5file['hist_info'].attrs['logscale'] = self.logscale
            hdf5file['hist_info'].attrs['normalize'] = self.normalize

            util.save_dataset(hdf5file, 'centers_x', self.centers_x)
            util.save_dataset(hdf5file, 'centers_y', self.centers_y)
            data = self.hist2d
            name = 'hist2d'
            attrs = {}
            if hasattr(data, 'unit'):
                hdf5file.create_dataset(name, data=data.value)
                attrs.update({'unit': self.hist2d.unit.to_string()})
            else:
                hdf5file.create_dataset(name, data=data)
            if self.colorlabel is not None:
                attrs.update({'colorlabel': self.colorlabel})
            else:
                print(("Unable to save colorlabel, please call "
                      + "the 'get_colorlabel' method before saving."))
            # Add attributes
            for key, attr in attrs.items():
                hdf5file[name].attrs[key] = attr

        util._copy_over_snapshot_information(self.snap, filename)



"""
This defines a base class for image creators (such as Projector and Slicer)
"""

import numpy as np
import warnings
from .. import util
from .. import settings
from .. import units
from ..orientation import Orientation


class ImageCreator:
    """
    This is a base class for creating images from a snapshot.
    """

    def __init__(self, snap, center, widths, direction, npix=512, parttype=0):
        """
        Initialize the ImageCreator class. This method will be called
        by subclasses such as Projector or Slicer.

        Parameters:
            snap (object): Snapshot object from which the image is created.

            center: Center of the image (3D coordinates).

            widths: Width of the image in each direction (3D coordinates)

            direction (str, Orientation): Direction of the image ('x', 'y', 'z')
                or an Orientation instance.

            npix (int): Number of pixels in the image (default is 512).
        """

        self._obervers = []

        self.snap = snap

        if settings.use_units:
            code_length = self.snap.length

        if hasattr(center, 'unit'):
            self._center = center.copy
            assert center.unit == code_length.unit, 'this restriction applies'
        elif settings.use_units:
            self._center = np.array(center) * code_length
        else:
            self._center = np.array(center)

        # Difference from original center of image
        if settings.use_units:
            self._diff_center = self._center.copy * 0
        else:
            self._diff_center = np.array(self._center) * 0

        if hasattr(widths, 'unit'):
            self._widths = widths.copy
            assert widths.unit == code_length.unit, 'this restriction applies'
        elif settings.use_units:
            self._widths = np.array(widths) * code_length
        else:
            self._widths = np.array(widths)

        if isinstance(direction, str):

            self.direction = direction
            if direction == 'x':
                self.orientation = Orientation(normal_vector=[1, 0, 0],
                                               perp_vector1=[0, 1, 0])
            elif direction == 'y':
                warnings.warn("\n\nThe definition of a 'y'-image was changed "
                              + "on 19/1/24 to be consistent with a right-handed "
                              + "coordinate system. You can recover the old "
                              + "behaviour by passing an orientation object:\n"
                              + "Orientation(normal_vector=[0, -1, 0], perp_vector1=[1, 0, 0])\n"
                              + "instead of a direction 'y' string.\n\n")
                self.orientation = Orientation(normal_vector=[0, 1, 0],
                                               perp_vector1=[0, 0, 1])
            elif direction == 'z':
                self.orientation = Orientation(normal_vector=[0, 0, 1],
                                               perp_vector1=[1, 0, 0])

        elif isinstance(direction, Orientation):
            self.direction = 'orientation'
            self.orientation = direction

        self._npix = self._npix_width = npix

        self._parttype = parttype

        # For checking if properties changes
        self._old_orientation = self.orientation.copy
        self._properties_changed = False

        # Check that units are consistent
        self.do_unit_consistency_check()

        # Check if image region exceeds simulation domain
        # TODO

    def _check_if_properties_changed(self):

        # Check if orientation changed compared to last
        # time _do_region_selection was called
        orientation_changed = not self.orientation._are_equal(self._old_orientation)

        if orientation_changed:
            self.direction == 'orientation'

        # print('_check_if_properties_changed called')
        if self._properties_changed or orientation_changed:
            self._do_region_selection()
            self._properties_changed = False
            self._old_orientation = self.orientation.copy
            return True
        else:
            return False

    def do_unit_consistency_check(self):
        """
        This function is called inside image creators to ensure
        that length units are consistent.
        """

        if settings.use_units:
            if hasattr(self, 'hsml'):
                assert self.hsml.unit.is_equivalent(self.widths.unit)
            if hasattr(self, 'pos'):
                assert self.pos.unit == self.widths.unit, (self.pos.unit, self.widths.unit)
            assert self.center.unit == self.widths.unit, (self.center.unit, self.widths.unit)
            if f"{self.parttype}_Coordinates" in self.snap:
                assert self.center.unit == self.snap[f"{self.parttype}_Coordinates"].unit
            if hasattr(self.snap, 'box_size'):
                assert self.center.unit == self.snap.box_size.unit

    def _do_region_selection(self):
        err_msg = ("_do_region_selection was called from ImageCreator. "
                   + "This should never happen as the Subclasses of "
                   + "ImageCreator should implement this method")
        raise RuntimeError(err_msg)

    def info(self):
        """
        Prints some information about the image creator instance.
        TODO: Add things like the number of cells/particles
        in the current instance.
        """
        s = f'Paicos ImageCreator instance of type: {str(self)}\n'
        s += f'widths: {self.widths}\n'
        s += f'center:  {self.center}\n'
        s += f'orientation:  {self.orientation}'
        print(s)

    @property
    def center(self):
        """
        Center of the image (3D coordinates).
        """
        return self._center

    @property
    def widths(self):
        """
        Width of the image in each direction (3D coordinates).
        """
        return self._widths

    @widths.setter
    def widths(self, value):
        """
        TODO: ensure that nx, ny
        remain consistent with widths
        """
        assert value.shape[0] == 3
        if settings.use_units:
            assert hasattr(value, 'unit')
            self._widths = value.copy
        else:
            self._widths = np.array(value)
        self._properties_changed = True

    @center.setter
    def center(self, value):
        assert value.shape[0] == 3
        if settings.use_units:
            assert hasattr(value, 'unit')
            self._center = value.copy
        else:
            self._center = np.array(value)
        self._properties_changed = True

    @property
    def npix(self):
        """
        The number of pixels along the horizontal
        direction of the image.
        """
        return self._npix

    @npix.setter
    def npix(self, value):
        self._npix = value
        # TODO: Not always needed to do this...
        self._properties_changed = True

    @property
    def npix_width(self):
        """
        The number of pixels along the horizontal
        direction of the image.
        """
        return self._npix

    @npix_width.setter
    def npix_width(self, value):
        self.npix = value

    @property
    def parttype(self):
        """
        The parttype being imaged, e.g. parttype=0
        are the Voronoi gas cells.
        """
        return self._parttype

    @property
    def x_c(self):
        """
        The x-component of .center
        """
        return self.center[0]

    @property
    def y_c(self):
        """
        The y-component of .center
        """
        return self.center[1]

    @property
    def z_c(self):
        """
        The z-component of .center
        """
        return self.center[2]

    @property
    def width_x(self):
        """
        The x-component of .widths
        """
        return self.widths[0]

    @property
    def width_y(self):
        """
        The y-component of .widths
        """
        return self.widths[1]

    @property
    def width_z(self):
        """
        The z-component of .widths
        """
        return self.widths[2]

    @property
    def extent(self):
        """
        The extent of the image in the horizontal-vertical plane.
        The idea with this property is that it corresponds to the
        matplotlib imshow keyword argument 'extent'. When using units,
        one has to remember to pass image_creator.extent.value rather
        than image_creator.extent.
        """
        if self.direction == 'orientation':

            # This is the extent of the image-plane pre-rotation
            # into the orientation of the image...
            # Use centered_extent for showing rotated images.
            center_width = self.x_c
            center_height = self.y_c

        elif self.direction == 'x':
            center_width = self.y_c
            center_height = self.z_c

        elif self.direction == 'y':
            center_width = self.z_c
            center_height = self.x_c

        elif self.direction == 'z':
            center_width = self.x_c
            center_height = self.y_c

        extent = [center_width - self.width / 2, center_width + self.width / 2,
                  center_height - self.height / 2, center_height + self.height / 2]

        if settings.use_units:
            extent = units.paicos_quantity_list_to_array(extent)
        else:
            extent = np.array(extent)
        return extent

    @property
    def centered_extent(self):
        """
        Same as 'extent' but centered on the center of the image.
        """
        centered_extent = [- self.width / 2, self.width / 2,
                           - self.height / 2, self.height / 2]

        if settings.use_units:
            centered_extent = units.paicos_quantity_list_to_array(centered_extent)
        else:
            centered_extent = np.array(centered_extent)
        return centered_extent

    @property
    def width(self):
        """
        The horizontal width of the image.
        """
        if self.direction == 'x':
            return self.width_y

        elif self.direction == 'y':
            return self.width_z

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_x

    @property
    def height(self):
        """
        The vertical height of the image.
        """
        if self.direction == 'x':
            return self.width_z

        elif self.direction == 'y':
            return self.width_x

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_y

    @property
    def depth(self):
        """
        The depth of the image.
        """
        if self.direction == 'x':
            return self.width_x

        elif self.direction == 'y':
            return self.width_y

        elif self.direction == 'z' or self.direction == 'orientation':
            return self.width_z

    @width.setter
    def width(self, value):
        if settings.use_units:
            assert hasattr(value, 'unit')
        if self.direction == 'x':
            self._widths[1] = value.copy

        elif self.direction == 'y':
            self._widths[2] = value.copy

        elif self.direction == 'z' or self.direction == 'orientation':
            self._widths[0] = value.copy
        self._properties_changed = True

    @height.setter
    def height(self, value):
        if settings.use_units:
            assert hasattr(value, 'unit')
        if self.direction == 'x':
            self._widths[2] = value.copy

        elif self.direction == 'y':
            self._widths[0] = value.copy

        elif self.direction == 'z' or self.direction == 'orientation':
            self._widths[1] = value.copy
        self._properties_changed = True

    @depth.setter
    def depth(self, value):
        if settings.use_units:
            assert hasattr(value, 'unit')
        if self.direction == 'x':
            self._widths[0] = value.copy

        elif self.direction == 'y':
            self._widths[1] = value.copy

        elif self.direction == 'z' or self.direction == 'orientation':
            self._widths[2] = value.copy
        self._properties_changed = True

    def double_resolution(self):
        """
        Calling this method doubles the pixel resolution of the image
        creator.
        """
        self.npix = self.npix * 2

    def half_resolution(self):
        """
        Calling this method halves the pixel resolution of the image
        creator.
        """
        self.npix = self.npix // 2

    def zoom(self, factor):
        """
        Calling this method zooms the view by the input factor,
        that is, the horizontal width and vertical height of the
        image creator is changed like this:

        width = width / factor

        height = height / factor

        A factor less than one zooms out the view.
        """
        self.width = self.width / factor
        self.height = self.height / factor

        self._properties_changed = True

    def _move_center_along_unit_vector(self, shift, unit_vector):
        """
        Helper function for moving center along unit vector.
        """
        if settings.use_units:
            assert hasattr(shift, 'unit')

        # Calculate new center
        new_center = self._center + shift * unit_vector
        # Keep track of changes to center (offset from initialized center)
        self._diff_center += new_center - self._center

        # Store new center and set bool for re-calculation of image
        self._center = new_center
        self._properties_changed = True

    def move_center_along_normal_vector(self, shift):
        """
        Moves the center of the image along its depth direction.
        Value can be positive or negative and must have same units
        as the center.

        (TODO: In principle we could just convert the distance to same units).
        """
        self._move_center_along_unit_vector(shift, self.orientation.normal_vector)

    def move_center_along_perp_vector1(self, shift):
        """
        Moves the center of the image along its horizontal direction.
        Value can be positive or negative and must have same units
        as the center.

        (TODO: In principle we could just convert the distance to same units).
        """
        self._move_center_along_unit_vector(shift, self.orientation.perp_vector1)

    def move_center_along_perp_vector2(self, shift):
        """
        Moves the center of the image along its vertical direction.
        Value can be positive or negative and must have same units
        as the center.

        (TODO: In principle we could just convert the distance to same units).
        """
        self._move_center_along_unit_vector(shift, self.orientation.perp_vector2)

    @property
    def npix_height(self):
        """
        The number of pixels of an image in the vertical direction.
        """
        if settings.use_units:
            return int((self.height / self.width).value * self.npix_width)
        else:
            return int((self.height / self.width) * self.npix_width)

    @property
    def area(self):
        """
        The area of the image plane.
        """
        return (self.extent[1] - self.extent[0]) * (self.extent[3] - self.extent[2])

    @property
    def area_per_pixel(self):
        """
        The area per pixel of the image plane.
        """
        return self.area / (self.npix_width * self.npix_height)

    @property
    def volume(self):
        """
        The volume of the projection region (0 for slices).
        """
        return self.width * self.height * self.depth

    @property
    def volume_per_pixel(self):
        """
        The volume per pixel in the image.
        """
        return self.volume / (self.npix_width * self.npix_height)



import numba
from numba import cuda
import math
import cupy as cp

# Hardcoded for uint64 coordinates (don't change without also changing morton
# code functions)
L = 21


@cuda.jit(device=True, inline=True)
def part1by2_64(n):
    n &= 0x1fffff
    n = (n | (n << 32)) & 0x1f00000000ffff
    n = (n | (n << 16)) & 0x1f0000ff0000ff
    n = (n | (n << 8)) & 0x100f00f00f00f00f
    n = (n | (n << 4)) & 0x10c30c30c30c30c3
    n = (n | (n << 2)) & 0x1249249249249249
    return n


@cuda.jit(device=True, inline=True)
def unpart1by2_64(n):
    n &= 0x1249249249249249
    n = (n ^ (n >> 2)) & 0x10c30c30c30c30c3
    n = (n ^ (n >> 4)) & 0x100f00f00f00f00f
    n = (n ^ (n >> 8)) & 0x1f0000ff0000ff
    n = (n ^ (n >> 16)) & 0x1f00000000ffff
    n = (n ^ (n >> 32)) & 0x1fffff
    return n


@cuda.jit(device=True, inline=True)
def encode64(x, y, z):
    xx = part1by2_64(x)
    yy = part1by2_64(y)
    zz = part1by2_64(z)
    return xx * 4 + yy * 2 + zz


@cuda.jit(device=True, inline=True)
def decode64(key):
    x = unpart1by2_64(key >> 2)
    y = unpart1by2_64(key >> 1)
    z = unpart1by2_64(key)
    return numba.float64(x), numba.float64(y), numba.float64(z)


@cuda.jit
def get_morton_keys64(pos, morton_keys):
    ip = cuda.grid(1)
    if ip < pos.shape[0]:
        morton_keys[ip] = encode64(pos[ip, 0], pos[ip, 1], pos[ip, 2])


@cuda.jit
def decode_morton_keys64(morton_keys, x, y, z):
    ip = cuda.grid(1)
    if ip < x.shape[0]:
        x[ip], y[ip], z[ip] = decode64(morton_keys[ip])


@cuda.jit(device=True, inline=True)
def device_clf_64(tmp):
    # This is
    # https://docs.nvidia.com/cuda/'libdevice-users-guide/__nv_clzll.html#__nv_clzll
    return cuda.libdevice.clzll(tmp)


@cuda.jit(device=True, inline=True)
def findSplit(sortedMortonCodes, first, last):
    # Identical Morton codes => split the range in the middle.

    firstCode = sortedMortonCodes[first]
    lastCode = sortedMortonCodes[last]

    if (firstCode == lastCode):
        return (first + last) >> 1

    # Calculate the number of highest bits that are the same
    # for all objects, using the count-leading-zeros intrinsic.

    commonPrefix = cuda.libdevice.clzll(firstCode ^ lastCode)

    # Use binary search to find where the next bit differs.
    # Specifically, we are looking for the highest object that
    # shares more than commonPrefix bits with the first one.

    split = first  # initial guess
    step = last - first

    while (step > 1):
        step = (step + 1) >> 1  # exponential decrease
        newSplit = split + step  # proposed new position

        if (newSplit < last):
            splitCode = sortedMortonCodes[newSplit]
            splitPrefix = cuda.libdevice.clzll(firstCode ^ splitCode)
            if (splitPrefix > commonPrefix):
                split = newSplit  # accept proposal

    return split


@cuda.jit()
def generateHierarchy(sortedMortonCodes, tree_children, tree_parents):
    """
    Generate tree using the sorted Morton code.
    This is done in parallel, see this blogpost
    https://developer.nvidia.com/blog/thinking-parallel-part-iii-tree-construction-gpu/
    and the corresponding paper:

    """
    n_codes = sortedMortonCodes.shape[0]
    num_internal_nodes = n_codes - 1
    idx = cuda.grid(1)

    # Launch thread for each internal node
    if idx < num_internal_nodes:

        # Find out which range of objects the node corresponds to.
        # (This is where the magic happens!)

        # Determine range
        first, last = determineRange(sortedMortonCodes, n_codes, idx)

        # Determine where to split the range.
        split = findSplit(sortedMortonCodes, first, last)

        # Select childA.
        if (split == first):
            # Is a leaf
            childA = split + num_internal_nodes
        else:
            # Is in an internal node
            childA = split

        # Select childB.
        if (split + 1 == last):
            childB = split + 1 + num_internal_nodes
        else:
            childB = split + 1

        # Record parent-child relationships.
        tree_children[idx, 0] = childA
        tree_children[idx, 1] = childB
        tree_parents[childA] = idx
        tree_parents[childB] = idx
    cuda.syncthreads()


@cuda.jit(device=True, inline=True)
def determineRange(sortedMortonCodes, n_codes, idx):
    """
    The determine range function needed by the generateHierarchy function.
    This code has has been adapted from the CornerStone CUDA/C++ code,
    see constructInternalNode in btree.hpp.
    """
    d = 1
    minPrefixLength = -1
    firstIndex = idx
    max_last = n_codes

    if (firstIndex > 0):
        p1 = sortedMortonCodes[firstIndex] ^ sortedMortonCodes[firstIndex + 1]
        m1 = sortedMortonCodes[firstIndex] ^ sortedMortonCodes[firstIndex - 1]

        # Count leading zeros
        lz_p1 = device_clf_64(p1)
        lz_m1 = device_clf_64(m1)

        if lz_p1 > lz_m1:
            d = 1
            minPrefixLength = lz_m1
        else:
            d = -1
            minPrefixLength = lz_p1

    searchRange = 2
    secondIndex = firstIndex + searchRange * d

    while (0 <= secondIndex and secondIndex < max_last):
        lz_first_second = device_clf_64(
            sortedMortonCodes[firstIndex] ^ sortedMortonCodes[secondIndex])
        if lz_first_second > minPrefixLength:
            searchRange *= 2
            secondIndex = firstIndex + searchRange * d
            if secondIndex < max_last:
                lz_first_second = device_clf_64(
                    sortedMortonCodes[firstIndex] ^ sortedMortonCodes[secondIndex])
            else:
                break
        else:
            break

    # start binary search with known searchRange
    secondIndex = firstIndex
    while searchRange > 1:
        searchRange = (searchRange + 1) // 2
        newJdx = secondIndex + searchRange * d
        if (0 <= newJdx and newJdx < max_last):
            lz_f_Jdx = device_clf_64(
                sortedMortonCodes[firstIndex] ^ sortedMortonCodes[newJdx])
            if lz_f_Jdx > minPrefixLength:
                secondIndex = newJdx

    first = min(secondIndex, firstIndex)
    last = max(secondIndex, firstIndex)

    return first, last


@cuda.jit()
def set_leaf_bounding_volumes(tree_bounds, points, half_size, conversion_factor):
    num_internal_nodes = half_size.shape[0] - 1
    num_leafs = half_size.shape[0]
    f = conversion_factor
    ip = cuda.grid(1)
    if ip < num_leafs:
        x_min = max(math.floor((points[ip, 0] - f * half_size[ip])), 0)
        y_min = max(math.floor((points[ip, 1] - f * half_size[ip])), 0)
        z_min = max(math.floor((points[ip, 2] - f * half_size[ip])), 0)
        x_max = min(math.ceil((points[ip, 0] + f * half_size[ip])), 2**L - 1)
        y_max = min(math.ceil((points[ip, 1] + f * half_size[ip])), 2**L - 1)
        z_max = min(math.ceil((points[ip, 2] + f * half_size[ip])), 2**L - 1)
        tree_bounds[ip + num_internal_nodes, 0, 0] = numba.uint64(x_min)
        tree_bounds[ip + num_internal_nodes, 1, 0] = numba.uint64(y_min)
        tree_bounds[ip + num_internal_nodes, 2, 0] = numba.uint64(z_min)
        tree_bounds[ip + num_internal_nodes, 0, 1] = numba.uint64(x_max)
        tree_bounds[ip + num_internal_nodes, 1, 1] = numba.uint64(y_max)
        tree_bounds[ip + num_internal_nodes, 2, 1] = numba.uint64(z_max)


@cuda.jit
def propagate_bounds_upwards(tree_bounds, tree_parents, num_leafs, num_internal_nodes):
    ip = cuda.grid(1)
    # Launch tread for each *leaf*
    if ip < num_leafs:
        child = ip + num_internal_nodes
        next_parent = tree_parents[child]

        while next_parent != -1:  # indicates hitting root
            x_min = tree_bounds[child, 0, 0]
            y_min = tree_bounds[child, 1, 0]
            z_min = tree_bounds[child, 2, 0]
            x_max = tree_bounds[child, 0, 1]
            y_max = tree_bounds[child, 1, 1]
            z_max = tree_bounds[child, 2, 1]
            cuda.atomic.min(tree_bounds, (next_parent, 0, 0), x_min)
            cuda.atomic.min(tree_bounds, (next_parent, 1, 0), y_min)
            cuda.atomic.min(tree_bounds, (next_parent, 2, 0), z_min)
            cuda.atomic.max(tree_bounds, (next_parent, 0, 1), x_max)
            cuda.atomic.max(tree_bounds, (next_parent, 1, 1), y_max)
            cuda.atomic.max(tree_bounds, (next_parent, 2, 1), z_max)

            # # This hopefully kills off one of the children threads
            # if child_min_key != tree_bounds[next_parent, 0]:
            #     break

            # Prepare next iteration of while loop
            child = next_parent
            next_parent = tree_parents[next_parent]


@cuda.jit(device=True, inline=True)
def distance(point, query_point):
    dist = math.sqrt((point[0] - query_point[0])**2
                     + (point[1] - query_point[1])**2
                     + (point[2] - query_point[2])**2)
    return dist


@cuda.jit(device=True, inline=True)
def is_point_in_box(point, box):
    for ii in range(3):
        if point[ii] < box[ii, 0] or point[ii] > box[ii, 1]:
            return False
    return True


@cuda.jit(device=True, inline=True)
def nearest_neighbor_device(points, tree_parents, tree_children, tree_bounds,
                            query_point, num_internal_nodes):

    # We traverse the nodes and leafs using a while loop and a queue.
    # Local memory on each tread (32 should be fine?)
    queue = cuda.local.array(128, numba.int64)
    # Initialize queue_index an start at node 0
    queue_index = 0
    queue[queue_index] = 0

    # Initialize min_dist, and min_index
    min_dist = (2**L - 1.0)
    min_index = -1

    while queue_index >= 0:

        node_id = queue[queue_index]
        # print(queue_index, node_id)#traverseA, traverseB, is_leafA, is_leafB)

        childA = tree_children[node_id, 0]
        childB = tree_children[node_id, 1]

        is_leafA = childA >= num_internal_nodes
        is_leafB = childB >= num_internal_nodes

        point_in_A = is_point_in_box(query_point, tree_bounds[childA])
        point_in_B = is_point_in_box(query_point, tree_bounds[childB])

        # Do explicit check if in a leaf
        if point_in_A and is_leafA:
            data_id = childA - num_internal_nodes
            dist = distance(points[data_id], query_point)

            if dist < min_dist:
                min_dist = dist
                min_index = data_id

        if point_in_B and is_leafB:
            data_id = childB - num_internal_nodes
            dist = distance(points[data_id], query_point)

            if dist < min_dist:
                min_dist = dist
                min_index = data_id

        # Whether to traverse
        traverseA = point_in_A and not is_leafA
        traverseB = point_in_B and not is_leafB

        if (not traverseA) and (not traverseB):
            queue_index -= 1
        else:
            if traverseA:
                queue[queue_index] = childA
            else:
                queue[queue_index] = childB
            if traverseA and traverseB:
                queue_index += 1
                queue[queue_index] = childB

    return min_dist, min_index


@cuda.jit
def find_nearest_neighbors(points, tree_parents, tree_children, tree_bounds,
                           query_points, dists, ids):
    ip = cuda.grid(1)
    n_queries = query_points.shape[0]
    if ip < n_queries:
        num_internal_nodes = numba.int64(tree_children.shape[0])

        # Select a query_point from the list of queries
        query_point = query_points[ip]

        min_dist, min_index = nearest_neighbor_device(points, tree_parents, tree_children,
                                                      tree_bounds, query_point,
                                                      num_internal_nodes)

        dists[ip] = min_dist
        ids[ip] = min_index


def get_blocks(size, threadsperblock):
    return (size + (threadsperblock - 1)) // threadsperblock


class GpuBinaryTree:
    def __init__(self, positions, sizes, threadsperblock=32):
        """
        Python/Numba implementation of a BVH (boundary volume hierarchy) tree.
        """

        # TODO: Check if positions and sizes are
        # CuPy arrays. If not, automatically transfer.
        # If they are Paicos Quantities, store their
        # units so that the distance calculation can be returned with units.

        # Copy positions
        self._pos = cp.array(positions)

        # Convert self._position to units from [0, 1) * 2**L
        self.off_sets = cp.min(self._pos, axis=0)
        self._pos -= self.off_sets[None, :]

        max_pos = cp.max(self._pos)
        self.conversion_factor = (2**L - 1) / cp.float64(max_pos)
        self._pos *= self.conversion_factor

        # Some tree properties
        self.num_leafs = self._pos.shape[0]
        self.num_internal_nodes = self.num_leafs - 1
        self.num_leafs_and_nodes = 2 * self.num_leafs - 1

        # Find blocks per grid
        blocks_leafs = get_blocks(self.num_leafs, threadsperblock)
        blocks_nodes = get_blocks(self.num_internal_nodes, threadsperblock)

        # Calculate uint64 and Morton keys
        self._pos_uint = self._pos.astype(cp.uint64)
        self.morton_keys = cp.zeros(self.num_leafs, dtype=cp.uint64)
        get_morton_keys64[blocks_leafs, threadsperblock](
            self._pos_uint, self.morton_keys)

        # TODO: delete integer positions? Not really needed after this step

        # Store index for going back to original ids.
        self.sort_index = cp.argsort(self.morton_keys)

        # Do the sorting
        self.morton_keys = self.morton_keys[self.sort_index]
        self._pos = self._pos[self.sort_index, :]
        self._pos_uint = self._pos_uint[self.sort_index, :]

        # Allocate arrays
        self.children = -1 * cp.ones((self.num_internal_nodes, 2), dtype=int)
        self.parents = -1 * cp.ones(self.num_leafs_and_nodes, dtype=int)

        # This sets the parent and children properties
        generateHierarchy[blocks_nodes, threadsperblock](self.morton_keys, self.children,
                                                         self.parents)

        # Set the boundaries for the leafs
        self.bounds = cp.zeros(
            (self.num_leafs_and_nodes, 3, 2), dtype=cp.uint64)

        # Ensures that atomic min and max does not just set to starting values
        self.bounds[:, :, 0] = 2**(3 * L) - 1
        self.bounds[:, :, 1] = 0

        set_leaf_bounding_volumes[blocks_leafs, threadsperblock](self.bounds, self._pos,
                                                                 sizes[self.sort_index],
                                                                 self.conversion_factor)

        # Calculate the bounding volumes for internal nodes,
        # by propagating the information upwards in the tree
        propagate_bounds_upwards[blocks_leafs, threadsperblock](self.bounds, self.parents,
                                                                self.num_leafs, self.num_internal_nodes)

    def _to_tree_coordinates(self, pos):
        return (pos - self.off_sets[None, :]) * self.conversion_factor

    def _tree_node_ids_to_data_ids(self, ids):
        """
        Go from sorted ids to original ordering of
        *selected* data.
        """
        return cp.arange(self.sort_index.shape[0])[self.sort_index][ids]

    def nearest_neighbor(self, query_points, threadsperblock=32):
        # TODO: check whether query points are already on GPU

        if query_points.ndim == 1:
            n_queries = 1
        else:
            n_queries = query_points.shape[0]

        blocks = get_blocks(n_queries, threadsperblock)

        ids = cp.zeros(n_queries, dtype=cp.int64)
        dists = cp.zeros(n_queries, dtype=cp.float64)
        find_nearest_neighbors[blocks, threadsperblock](
            self._pos, self.parents, self.children, self.bounds,
            self._to_tree_coordinates(query_points), dists, ids)
        return dists / self.conversion_factor, self._tree_node_ids_to_data_ids(ids)


if __name__ == '__main__':
    import paicos as pa
    from scipy.spatial import KDTree
    import numpy as np
    pa.use_units(False)
    snap = pa.Snapshot(pa.data_dir + 'snap_247.hdf5')
    center = snap.Cat.Group['GroupPos'][0]
    widths = np.array([15000, 15000, 15000])

    index = pa.util.get_index_of_cubic_region_plus_thin_layer(
        snap['0_Coordinates'], center, widths,
        snap['0_Diameters'], snap.box)

    snap = snap.select(index, parttype=0)

    pos_cpu = snap['0_Coordinates']
    pos = cp.array(pos_cpu)
    sizes = cp.array(snap['0_Diameters'])
    bvh_tree = GpuBinaryTree(pos, 1.2 * sizes)

    bvh_dist, bvh_ids = bvh_tree.nearest_neighbor(pos)

    # Construct a scipy kd tree
    kd_tree = KDTree(pos_cpu)

    kd_dist, kd_ids = kd_tree.query(pos_cpu)

    np.testing.assert_array_equal(kd_ids, bvh_ids.get())

    # Random query positions
    pos2_cpu = np.random.rand(10000, 3) * widths[None, :]
    pos2_cpu += (center - widths / 2)[None, :]

    pos2 = cp.array(pos2_cpu)

    bvh_dist2, bvh_ids2 = bvh_tree.nearest_neighbor(pos2)
    kd_dist2, kd_ids2 = kd_tree.query(pos2_cpu)

    np.testing.assert_array_equal(kd_ids2, bvh_ids2.get())



from cython.parallel import prange
cimport openmp
import numpy as np
cimport numpy as np

ctypedef fused real_t:
    float
    double
    np.float32_t
    np.float64_t


def get_cube(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
             real_t sidelength_x, real_t sidelength_y,
             real_t sidelength_z, real_t boxsize, int numthreads):

    """
    This is a cython implementation of a selection function,
    which selects points inside a rectangular region ('cube' is a misnomer...).

    Users should not use this low-level function but instead use
    paicos.util.get_index_of_cubic_region

    Parameters:
        pos (array, (N,3)): positions
        xc (double): x-position of center
        yc (double): y-position of center
        zc (double): z-position of center
        sidelength_x (double): length of cube along x
        sidelength_y (double): length of cube along y
        sidelength_z (double): length of cube along z
        boxsize (double): size of simulation domain,
                           which for now is assumed to be cubic!
        numthreads (int): number of openmp threads to use

    Returns:
        boolean array (N): A boleean array with True for points inside the selected region
    """

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        # Index calculation
        index[ip] = 0
        if (x < sidelength_x/2.0) and (x > -sidelength_x/2.0):
            if (y < sidelength_y/2.0) and (y > -sidelength_y/2.0):
                if (z > -sidelength_z/2.0):
                    if (z < sidelength_z/2.0):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp


def get_cube_plus_thin_layer(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
                             real_t sidelength_x, real_t sidelength_y,
                             real_t sidelength_z, real_t [:] thickness,
                             real_t boxsize, int numthreads):
    """
    This is a cython implementation of a selection function,
    which selects points inside a rectangular region ('cube' is a misnomer...)
    + a layer of variable thickness.

    Users should not use this low-level function but instead use
    paicos.util.get_index_of_cubic_region_plus_thin_layer

    Parameters:
        pos (array, (N,3)): positions
        xc (double): x-position of center
        yc (double): y-position of center
        zc (double): z-position of center
        sidelength_x (double): length of cube along x
        sidelength_y (double): length of cube along y
        sidelength_z (double): length of cube along z
        thickness (array, (N)): variable thickness for each point
        boxsize (double): size of simulation domain,
                           which for now is assumed to be cubic!
        numthreads (int): number of openmp threads to use

    Returns:
        boolean array (N): A boleean array with True for points inside the selected region
    """

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        # Index calculation
        index[ip] = 0
        if (x < (sidelength_x/2.0 + thickness[ip])) and (x > -(sidelength_x/2.0 + thickness[ip])):
            if (y < (sidelength_y/2.0 + thickness[ip])) and (y > -(sidelength_y/2.0 + thickness[ip])):
                if (z < (sidelength_z/2.0 + thickness[ip])):
                    if (z > -(sidelength_z/2.0 + thickness[ip])):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp

def get_rotated_cube(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
                             real_t sidelength_x, real_t sidelength_y,
                             real_t sidelength_z,
                             real_t boxsize,
                             real_t[:] unit_vector_x,
                             real_t[:] unit_vector_y,
                             real_t[:] unit_vector_z,
                             int numthreads):
    """
    Same as get_cube but for a rotated cube.
    Please see paicos.util.get_index_of_rotated_cubic_region for details.
    """

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z
    cdef real_t x_dot_ex, x_dot_ey, x_dot_ez

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        x_dot_ex = x * unit_vector_x[0] + y * unit_vector_x[1] + z * unit_vector_x[2]
        x_dot_ey = x * unit_vector_y[0] + y * unit_vector_y[1] + z * unit_vector_y[2]
        x_dot_ez = x * unit_vector_z[0] + y * unit_vector_z[1] + z * unit_vector_z[2]

        # Index calculation
        index[ip] = 0
        if (x_dot_ex < sidelength_x/2.0) and (x_dot_ex > -sidelength_x/2.0):
            if (x_dot_ey < sidelength_y/2.0) and (x_dot_ey > -sidelength_y/2.0):
                if (x_dot_ez < sidelength_z/2.0):
                    if (x_dot_ez > -sidelength_z/2.0):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp

def get_rotated_cube_plus_thin_layer(real_t [:, :] pos, real_t xc, real_t yc, real_t zc,
                             real_t sidelength_x, real_t sidelength_y,
                             real_t sidelength_z, real_t [:] thickness,
                             real_t boxsize,
                             real_t[:] unit_vector_x,
                             real_t[:] unit_vector_y,
                             real_t[:] unit_vector_z,
                             int numthreads):
    """
    Same as get_cube_plus_thin_layer but for a rotated cube.
    Please see paicos.util.get_index_of_rotated_cubic_region_plus_thin_layer for details.
    """

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z
    cdef real_t x_dot_ex, x_dot_ey, x_dot_ez

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        # # box_wrap_diff function
        # if x < -0.5*boxsize:
        #     x = x + boxsize
        # elif x > 0.5*boxsize:
        #     x = x - boxsize

        # if y < -0.5*boxsize:
        #     y = y + boxsize
        # elif y > 0.5*boxsize:
        #     y = y - boxsize

        # if z < -0.5*boxsize:
        #     z = z + boxsize
        # elif z > 0.5*boxsize:
        #     z = z - boxsize

        x_dot_ex = x * unit_vector_x[0] + y * unit_vector_x[1] + z * unit_vector_x[2]
        x_dot_ey = x * unit_vector_y[0] + y * unit_vector_y[1] + z * unit_vector_y[2]
        x_dot_ez = x * unit_vector_z[0] + y * unit_vector_z[1] + z * unit_vector_z[2]

        # Index calculation
        index[ip] = 0
        if (x_dot_ex < (sidelength_x/2.0 + thickness[ip])) and (x_dot_ex > -(sidelength_x/2.0 + thickness[ip])):
            if (x_dot_ey < (sidelength_y/2.0 + thickness[ip])) and (x_dot_ey > -(sidelength_y/2.0 + thickness[ip])):
                if (x_dot_ez < (sidelength_z/2.0 + thickness[ip])):
                    if (x_dot_ez > -(sidelength_z/2.0 + thickness[ip])):
                        index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp


def get_radial_range(real_t [:, :] pos, real_t xc, real_t yc,
                    real_t zc, real_t r_min, real_t r_max,
                    int numthreads):
    """
    This is a cython implementation of a selection function,
    which selects points inside a spherical shell.

    Users should not use this low-level function but instead use
    paicos.util.get_index_of_radial_range

    Parameters:
        pos (array, (N,3)): positions
        xc (double): x-position of center
        yc (double): y-position of center
        zc (double): z-position of center
        r_min (double): mininum radius
        r_max (double): maxinum radius
        numthreads (int): number of openmp threads to use

    Returns:
        boolean array (N): A boleean array with True for points inside the selected region
    """

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z, r2

    cdef real_t r2_min = r_min*r_min
    cdef real_t r2_max = r_max*r_max

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        r2 = x*x + y*y + z*z

        # Index calculation
        index[ip] = 0
        if (r2 < r2_max) and (r2 > r2_min):
            index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp


def get_radial_range_plus_thin_layer(real_t [:, :] pos, real_t xc, real_t yc,
                                     real_t zc, real_t r_min, real_t r_max,
                                     real_t [:] thickness, int numthreads):
    """
    This is a cython implementation of a selection function,
    which selects points inside a spherical shell + a thin layer with
    variable thickness.

    Users should not use this low-level function but instead use
    paicos.util.get_index_of_radial_range_plus_thin_layer

    Parameters:
        pos (array, (N,3)): positions
        xc (double): x-position of center
        yc (double): y-position of center
        zc (double): z-position of center
        r_min (double): mininum radius
        r_max (double): maxinum radius
        numthreads (int): number of openmp threads to use

    Returns:
        boolean array (N): A boleean array with True for points inside the selected region
    """

    cdef int Np = pos.shape[0]
    cdef int ip
    cdef real_t x, y, z, r2

    # cdef real_t r2_min = r_min*r_min
    # cdef real_t r2_max = r_max*r_max
    cdef real_t r2_min, r2_max

    cdef int[:] index = np.zeros(Np, dtype=np.intc)

    openmp.omp_set_num_threads(numthreads)

    for ip in prange(Np, nogil=True, schedule='static'):
        x = pos[ip, 0] - xc
        y = pos[ip, 1] - yc
        z = pos[ip, 2] - zc

        r2 = x*x + y*y + z*z

        r2_min = 0.0
        if (r_min > thickness[ip]):
            r2_min = (r_min - thickness[ip])*(r_min - thickness[ip])
        r2_max = (r_max + thickness[ip])*(r_max + thickness[ip])
        
        # Index calculation
        index[ip] = 0
        if (r2 < r2_max) and (r2 > r2_min):
            index[ip] = 1

    # Return a numpy boolean array
    tmp = np.zeros(Np, dtype=np.bool_)
    tmp[:] = index[:]
    return tmp



"""This defines a reader for Arepo Friends-of-Friends (FoF) and subhalo catalog files"""
import numpy as np
import h5py
from .paicos_readers import PaicosReader
from ..writers.paicos_writer import PaicosWriter
from .. import settings
import numbers
import warnings


class Catalog(PaicosReader):
    """
    This is a Python class for reading Arepo group and subhalo catalogs.
    The class is based on a script originally written by Ewald Puchwein,
    which has since then been modified and included in Paicos.

    The class takes in the path of the directory containing the catalog, the
    catalog number, and an optional basename parameter, and uses this
    information to locate and open the catalog files. The class also loads
    the catalog's header, parameters, and configuration.
    The class also includes methods to extract the redshift, scale
    factor, and other properties of the catalog.

    Important methods and attributes:
    ---------------------------------

        cat = Catalog()

        cat.Group : dict
            Contains a dictionary of the FoF-catalog.

        cat.Sub : dict
            Contains a dictionary of the Subfind-catalog.

        cat.Parameters : dict
            Contains information from the parameter file used in the simulation (e.g. param.txt).

        cat.Config : dict
            Contains information from the Config file used in the simulation (e.g. Config.txt).

        cat.Header : dict
            Contains information about this particular catalog such as its time (e.g scale factor).

        cat.z : float
            The redshift.

        cat.h : float
            Reduced Hubble param (e.g. 0.67).

        cat.age : float
            The age of the Universe (only for cosmological runs).

        cat.lookback_time : float
            The age of the Universe (only for cosmological runs).

    """
    def __init__(self, basedir='.', snapnum=None, load_all=True,
                 to_physical=False, subfind_catalog=True, verbose=False):
        """
        Initializes the Catalog class.

        Parameters
        ----------

            basedir : str
                The path of the directory containing the catalogs
                (e.g. the 'output' folder).

            snapnum : int
                The snapshot number.

            load_all : bool
                Whether to immediately load all fields or not.

            to_physical : bool
                whether to convert to physical unit upon loading the data.
                Default is False.

            subfind_catalog : bool
                whether the simulation has subfind catalogs,
                when False the code will look for FoF-catalogs only.

            verbose : bool
                whether to print information, default is False.

        """

        if subfind_catalog:
            basename = 'fof_subhalo_tab'
        else:
            basename = 'fof_tab'

        super().__init__(basedir=basedir, snapnum=snapnum, basename=basename,
                         load_all=False, to_physical=to_physical,
                         basesubdir='groups', verbose=verbose)

        # give names to specific header fields
        self.nfiles = self.Header["NumFiles"]
        self.ngroups = self.Header["Ngroups_Total"]

        if "Nsubgroups_Total" in self.Header.keys():
            self.nsubs = self.Header["Nsubgroups_Total"]
        else:
            self.nsubs = self.Header["Nsubhalos_Total"]

        # Initialize dictionaries
        self.Group = {}
        self.Sub = {}

        # Load all data
        if load_all:
            self.load_all_data()

    def load_data(self):
        """
        Overwrite of the base class method.

        :meta private:
        """
        pass

    def load_all_data(self):
        """
        Calling this method simply loads all the data in the catalog.

        TODO: For large catalogs it might be useful to implement on-demand
        access in a similar way to what we have for snapshots.
        """

        skip_gr = 0
        skip_sub = 0
        for ifile in range(self.nfiles):
            if self.multi_file is False:
                cur_filename = self.filename
            else:
                if self.no_subdir:
                    cur_filename = self.multi_wo_dir.format(ifile)
                else:
                    cur_filename = self.multi_filename.format(ifile)

            if self.verbose:
                print("reading file", cur_filename)

            file = h5py.File(cur_filename, "r")

            ng = int(file["Header"].attrs["Ngroups_ThisFile"])

            if "Nsubgroups_ThisFile" in file["Header"].attrs.keys():
                ns = int(file["Header"].attrs["Nsubgroups_ThisFile"])
            else:
                ns = int(file["Header"].attrs["Nsubhalos_ThisFile"])

            # initialize arrays
            if ifile == 0:
                if "Group" in file:
                    for ikey in file["Group"].keys():
                        if len(file["Group/" + ikey].shape) == 1:
                            self.Group[ikey] = np.empty(
                                self.ngroups, dtype=file["Group/" + ikey].dtype)
                        elif len(file["Group/" + ikey].shape) == 2:
                            self.Group[ikey] = np.empty(
                                (self.ngroups, file["Group/" + ikey].shape[1]),
                                dtype=file["Group/" + ikey].dtype)
                        else:
                            assert False
                if "Subhalo" in file:
                    for ikey in file["Subhalo"].keys():
                        if len(file["Subhalo/" + ikey].shape) == 1:
                            self.Sub[ikey] = np.empty(
                                self.nsubs, dtype=file["Subhalo/" + ikey].dtype)
                        elif len(file["Subhalo/" + ikey].shape) == 2:
                            self.Sub[ikey] = np.empty(
                                (self.nsubs, file["Subhalo/" + ikey].shape[1]),
                                dtype=file["Subhalo/" + ikey].dtype)
                        else:
                            assert False

            # read group data
            if ng > 0:
                for ikey in file["Group"].keys():
                    self.Group[ikey][skip_gr:skip_gr + ng] = file["Group/" + ikey]

            # read subhalo data
            if ns > 0:
                for ikey in file["Subhalo"].keys():
                    self.Sub[ikey][skip_sub:skip_sub + ns] = file["Subhalo/" + ikey]

            skip_gr += ng
            skip_sub += ns

            file.close()

        # Load all variables with double precision
        if settings.double_precision:

            for ikey in self.Group:
                if not issubclass(self.Group[ikey].dtype.type, numbers.Integral):
                    self.Group[ikey] = self.Group[ikey].astype(np.float64)

            for ikey in self.Sub:
                if not issubclass(self.Sub[ikey].dtype.type, numbers.Integral):
                    self.Sub[ikey] = self.Sub[ikey].astype(np.float64)
        else:
            warnings.warn('\n\nThe cython routines expect double precision '
                          + 'and will fail unless settings.double_precision '
                          + 'is True.\n\n')

        if settings.use_units:
            for key in list(self.Group.keys()):
                self.Group[key] = self.get_paicos_quantity(
                    self.Group[key], key,
                    field='groups')
                if not hasattr(self.Group[key], 'unit'):
                    del self.Group[key]

            for key in list(self.Sub.keys()):
                self.Sub[key] = self.get_paicos_quantity(
                    self.Sub[key], key,
                    field='subhalos')
                if not hasattr(self.Sub[key], 'unit'):
                    del self.Sub[key]

    def save_new_catalog(self, basename, single_precision=False):
        """
        Save a new catalog containing only the currently loaded
        variables. Useful for reducing datasets to smaller sizes.
        """
        writer = PaicosWriter(self, self.basedir, basename, 'w')

        for key in self.Group:
            Ngroups_Total = self.Group[key].shape[0]

            if single_precision:
                data = self.Group[key].astype(np.float32)
            else:
                data = self.Group[key]
            writer.write_data(key, data, group='Group')

        for key in self.Sub:
            Nsubgroups_Total = self.Sub[key].shape[0]

            if single_precision:
                data = self.Sub[key].astype(np.float32)
            else:
                data = self.Sub[key]
            writer.write_data(key, data, group='Subhalo')

        with h5py.File(writer.tmp_filename, 'r+') as f:
            f['Header'].attrs["Ngroups_ThisFile"] = Ngroups_Total
            f['Header'].attrs["Ngroups_Total"] = Ngroups_Total
            f['Header'].attrs["Nsubgroups_ThisFile"] = Nsubgroups_Total
            f['Header'].attrs["Nsubgroups_Total"] = Nsubgroups_Total
            f['Header'].attrs["NumFiles"] = 1

        writer.finalize()

        return writer



import cython
import numpy as np

cdef extern int __builtin_clzl(unsigned long x)


def leading_zeros_cython(unsigned long x):
    return __builtin_clzl(x)

def get_leading_zeros(unsigned long[:] morton_keys):
    cdef int ii, N
    N = morton_keys.shape[0] - 1
    cdef int[:] my_leading_zeros = np.zeros(N, dtype=np.int32)
    for ii in range(N):
        key = morton_keys[ii] ^ morton_keys[ii+1]
        my_leading_zeros[ii] = __builtin_clzl(key)
    tmp = np.zeros(N, dtype=np.int32)
    tmp[:] = my_leading_zeros[:]
    return tmp
